@@ -1,7 +1,8 @@
-/*
+/**
  * unistr.c - Unicode string handling. Part of the Linux-NTFS project.
  *
- * Copyright (c) 2000-2004 Anton Altaparmakov
+ * Copyright (c) 2000-2006 Anton Altaparmakov
+ * Copyright (c) 2005-2007 Yura Pakhuchiy
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -19,7 +20,9 @@
  * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #ifdef HAVE_STDIO_H
 #include <stdio.h>
@@ -37,9 +40,12 @@
 #include <errno.h>
 #endif
 
+#include "attrib.h"
+#include "endians.h"
 #include "types.h"
 #include "unistr.h"
 #include "debug.h"
+#include "logging.h"
 
 /*
  * IMPORTANT
@@ -53,7 +59,8 @@
  * This is used by the name collation functions to quickly determine what
  * characters are (in)valid.
  */
-const u8 legal_ansi_char_array[0x40] = {
+#if 0
+static const u8 legal_ansi_char_array[0x40] = {
 	0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
 	0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
 
@@ -66,6 +73,7 @@ const u8 legal_ansi_char_array[0x40] = {
 	0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17,
 	0x17, 0x17, 0x04, 0x16, 0x18, 0x16, 0x18, 0x18,
 };
+#endif
 
 /**
  * ntfs_names_are_equal - compare two Unicode names for equality
@@ -118,20 +126,20 @@ BOOL ntfs_names_are_equal(const ntfschar *s1, size_t s1_len,
  */
 int ntfs_names_collate(const ntfschar *name1, const u32 name1_len,
 		const ntfschar *name2, const u32 name2_len,
-		const int err_val, const IGNORE_CASE_BOOL ic,
-		const ntfschar *upcase, const u32 upcase_len)
+		const int err_val __attribute__((unused)),
+		const IGNORE_CASE_BOOL ic, const ntfschar *upcase,
+		const u32 upcase_len)
 {
 	u32 cnt;
-	ntfschar c1, c2;
+	u16 c1, c2;
 
 #ifdef DEBUG
 	if (!name1 || !name2 || (ic && (!upcase || !upcase_len))) {
-		Dputs("ntfs_names_collate received NULL pointer!");
+		ntfs_log_debug("ntfs_names_collate received NULL pointer!\n");
 		exit(1);
 	}
 #endif
-	for (cnt = 0; cnt < min(name1_len, name2_len); ++cnt)
-	{
+	for (cnt = 0; cnt < min(name1_len, name2_len); ++cnt) {
 		c1 = le16_to_cpu(*name1);
 		name1++;
 		c2 = le16_to_cpu(*name2);
@@ -142,8 +150,10 @@ int ntfs_names_collate(const ntfschar *name1, const u32 name1_len,
 			if (c2 < upcase_len)
 				c2 = le16_to_cpu(upcase[c2]);
 		}
+#if 0
 		if (c1 < 64 && legal_ansi_char_array[c1] & 8)
 			return err_val;
+#endif
 		if (c1 < c2)
 			return -1;
 		if (c1 > c2)
@@ -154,9 +164,11 @@ int ntfs_names_collate(const ntfschar *name1, const u32 name1_len,
 	if (name1_len == name2_len)
 		return 0;
 	/* name1_len > name2_len */
+#if 0
 	c1 = le16_to_cpu(*name1);
 	if (c1 < 64 && legal_ansi_char_array[c1] & 8)
 		return err_val;
+#endif
 	return 1;
 }
 
@@ -176,12 +188,12 @@ int ntfs_names_collate(const ntfschar *name1, const u32 name1_len,
  */
 int ntfs_ucsncmp(const ntfschar *s1, const ntfschar *s2, size_t n)
 {
-	ntfschar c1, c2;
+	u16 c1, c2;
 	size_t i;
 
 #ifdef DEBUG
 	if (!s1 || !s2) {
-		Dputs("ntfs_wcsncmp() received NULL pointer!");
+		ntfs_log_debug("ntfs_wcsncmp() received NULL pointer!\n");
 		exit(1);
 	}
 #endif
@@ -219,12 +231,12 @@ int ntfs_ucsncmp(const ntfschar *s1, const ntfschar *s2, size_t n)
 int ntfs_ucsncasecmp(const ntfschar *s1, const ntfschar *s2, size_t n,
 		const ntfschar *upcase, const u32 upcase_size)
 {
-	ntfschar c1, c2;
+	u16 c1, c2;
 	size_t i;
 
 #ifdef DEBUG
 	if (!s1 || !s2 || !upcase) {
-		Dputs("ntfs_wcsncasecmp() received NULL pointer!");
+		ntfs_log_debug("ntfs_wcsncasecmp() received NULL pointer!\n");
 		exit(1);
 	}
 #endif
@@ -289,22 +301,30 @@ ntfschar *ntfs_ucsndup(const ntfschar *s, u32 maxlen)
 	u32 len;
 
 	len = ntfs_ucsnlen(s, maxlen);
-	dst = malloc((len + 1) * sizeof(ntfschar));
+	dst = ntfs_malloc((len + 1) * sizeof(ntfschar));
 	if (dst) {
 		memcpy(dst, s, len * sizeof(ntfschar));
-		dst[len] = cpu_to_le16(L'\0');
+		dst[len] = 0;
 	}
 	return dst;
 }
 
 /**
- * ntfs_name_upcase
+ * ntfs_name_upcase - Map an Unicode name to its uppercase equivalent
+ * @name:
+ * @name_len:
+ * @upcase:
+ * @upcase_len:
+ *
+ * Description...
+ *
+ * Returns:
  */
 void ntfs_name_upcase(ntfschar *name, u32 name_len, const ntfschar *upcase,
 		const u32 upcase_len)
 {
 	u32 i;
-	ntfschar u;
+	u16 u;
 
 	for (i = 0; i < name_len; i++)
 		if ((u = le16_to_cpu(name[i])) < upcase_len)
@@ -312,7 +332,14 @@ void ntfs_name_upcase(ntfschar *name, u32 name_len, const ntfschar *upcase,
 }
 
 /**
- * ntfs_file_value_upcase
+ * ntfs_file_value_upcase - Convert a filename to upper case
+ * @file_name_attr:
+ * @upcase:
+ * @upcase_len:
+ *
+ * Description...
+ *
+ * Returns:
  */
 void ntfs_file_value_upcase(FILE_NAME_ATTR *file_name_attr,
 		const ntfschar *upcase, const u32 upcase_len)
@@ -322,7 +349,17 @@ void ntfs_file_value_upcase(FILE_NAME_ATTR *file_name_attr,
 }
 
 /**
- * ntfs_file_values_compare
+ * ntfs_file_values_compare - Which of two filenames should be listed first
+ * @file_name_attr1:
+ * @file_name_attr2:
+ * @err_val:
+ * @ic:
+ * @upcase:
+ * @upcase_len:
+ *
+ * Description...
+ *
+ * Returns:
  */
 int ntfs_file_values_compare(const FILE_NAME_ATTR *file_name_attr1,
 		const FILE_NAME_ATTR *file_name_attr2,
@@ -384,7 +421,7 @@ int ntfs_ucstombs(const ntfschar *ins, const int ins_len, char **outs,
 	}
 	if (!mbs) {
 		mbs_len = (ins_len + 1) * MB_CUR_MAX;
-		mbs = (char*)malloc(mbs_len);
+		mbs = (char*)ntfs_malloc(mbs_len);
 		if (!mbs)
 			return -1;
 	}
@@ -401,7 +438,7 @@ int ntfs_ucstombs(const ntfschar *ins, const int ins_len, char **outs,
 				errno = ENAMETOOLONG;
 				return -1;
 			}
-			tc = (char*)malloc((mbs_len + 64) & ~63);
+			tc = (char*)ntfs_malloc((mbs_len + 64) & ~63);
 			if (!tc)
 				goto err_out;
 			memcpy(tc, mbs, mbs_len);
@@ -422,7 +459,7 @@ int ntfs_ucstombs(const ntfschar *ins, const int ins_len, char **outs,
 		if (cnt == -1)
 			goto err_out;
 		if (cnt <= 0) {
-			Dprintf("Eeek. cnt <= 0, cnt = %i\n", cnt);
+			ntfs_log_debug("Eeek. cnt <= 0, cnt = %i\n", cnt);
 			errno = EINVAL;
 			goto err_out;
 		}
@@ -431,22 +468,19 @@ int ntfs_ucstombs(const ntfschar *ins, const int ins_len, char **outs,
 #ifdef HAVE_MBSINIT
 	/* Make sure we are back in the initial state. */
 	if (!mbsinit(&mbstate)) {
-		Dputs("Eeek. mbstate not in initial state!");
+		ntfs_log_debug("Eeek. mbstate not in initial state!\n");
 		errno = EILSEQ;
 		goto err_out;
 	}
 #endif
 	/* Now write the NULL character. */
-	mbs[o] = '\0';
+	mbs[o] = 0;
 	if (*outs != mbs)
 		*outs = mbs;
 	return o;
 err_out:
-	if (mbs != *outs) {
-		int eo = errno;
+	if (mbs != *outs)
 		free(mbs);
-		errno = eo;
-	}
 	return -1;
 }
 
@@ -527,7 +561,7 @@ int ntfs_mbstoucs(const char *ins, ntfschar **outs, int outs_len)
 	ins_len++;
 	if (!ucs) {
 		ucs_len = ins_len;
-		ucs = (ntfschar*)malloc(ucs_len * sizeof(ntfschar));
+		ucs = (ntfschar*)ntfs_malloc(ucs_len * sizeof(ntfschar));
 		if (!ucs)
 			return -1;
 	}
@@ -566,7 +600,7 @@ int ntfs_mbstoucs(const char *ins, ntfschar **outs, int outs_len)
 		if (cnt == -1)
 			goto err_out;
 		if (cnt < -1) {
-			Dprintf("%s(): Eeek. cnt = %i\n", __FUNCTION__, cnt);
+			ntfs_log_trace("Eeek. cnt = %i\n", cnt);
 			errno = EINVAL;
 			goto err_out;
 		}
@@ -582,23 +616,19 @@ int ntfs_mbstoucs(const char *ins, ntfschar **outs, int outs_len)
 #ifdef HAVE_MBSINIT
 	/* Make sure we are back in the initial state. */
 	if (!mbsinit(&mbstate)) {
-		Dprintf("%s(): Eeek. mbstate not in initial state!\n",
-				__FUNCTION__);
+		ntfs_log_trace("Eeek. mbstate not in initial state!\n");
 		errno = EILSEQ;
 		goto err_out;
 	}
 #endif
 	/* Now write the NULL character. */
-	ucs[o] = cpu_to_le16(L'\0');
+	ucs[o] = 0;
 	if (*outs != ucs)
 		*outs = ucs;
 	return o;
 err_out:
-	if (ucs != *outs) {
-		int eo = errno;
+	if (ucs != *outs)
 		free(ucs);
-		errno = eo;
-	}
 	return -1;
 }
 
@@ -610,60 +640,136 @@ err_out:
  * ntfs_upcase_table_build() builds the default upcase table for NTFS and
  * stores it in the caller supplied buffer @uc of size @uc_len.
  *
+ * The generated $UpCase table is the one used by Windows Vista.
+ *
  * Note, @uc_len must be at least 128kiB in size or bad things will happen!
  */
 void ntfs_upcase_table_build(ntfschar *uc, u32 uc_len)
 {
-	static int uc_run_table[][3] = { /* Start, End, Add */
-	{0x0061, 0x007B,  -32}, {0x0451, 0x045D, -80}, {0x1F70, 0x1F72,  74},
-	{0x00E0, 0x00F7,  -32}, {0x045E, 0x0460, -80}, {0x1F72, 0x1F76,  86},
-	{0x00F8, 0x00FF,  -32}, {0x0561, 0x0587, -48}, {0x1F76, 0x1F78, 100},
-	{0x0256, 0x0258, -205}, {0x1F00, 0x1F08,   8}, {0x1F78, 0x1F7A, 128},
-	{0x028A, 0x028C, -217}, {0x1F10, 0x1F16,   8}, {0x1F7A, 0x1F7C, 112},
-	{0x03AC, 0x03AD,  -38}, {0x1F20, 0x1F28,   8}, {0x1F7C, 0x1F7E, 126},
-	{0x03AD, 0x03B0,  -37}, {0x1F30, 0x1F38,   8}, {0x1FB0, 0x1FB2,   8},
-	{0x03B1, 0x03C2,  -32}, {0x1F40, 0x1F46,   8}, {0x1FD0, 0x1FD2,   8},
-	{0x03C2, 0x03C3,  -31}, {0x1F51, 0x1F52,   8}, {0x1FE0, 0x1FE2,   8},
-	{0x03C3, 0x03CC,  -32}, {0x1F53, 0x1F54,   8}, {0x1FE5, 0x1FE6,   7},
-	{0x03CC, 0x03CD,  -64}, {0x1F55, 0x1F56,   8}, {0x2170, 0x2180, -16},
-	{0x03CD, 0x03CF,  -63}, {0x1F57, 0x1F58,   8}, {0x24D0, 0x24EA, -26},
-	{0x0430, 0x0450,  -32}, {0x1F60, 0x1F68,   8}, {0xFF41, 0xFF5B, -32},
-	{0}
+	/*
+	 * "Start" is inclusive and "End" is exclusive, every value has the
+	 * value of "Add" added to it.
+	 */
+	static int add[][3] = { /* Start, End, Add */
+	{0x0061, 0x007b,   -32}, {0x00e0, 0x00f7,  -32}, {0x00f8, 0x00ff, -32}, 
+	{0x0256, 0x0258,  -205}, {0x028a, 0x028c, -217}, {0x037b, 0x037e, 130}, 
+	{0x03ac, 0x03ad,   -38}, {0x03ad, 0x03b0,  -37}, {0x03b1, 0x03c2, -32},
+	{0x03c2, 0x03c3,   -31}, {0x03c3, 0x03cc,  -32}, {0x03cc, 0x03cd, -64},
+	{0x03cd, 0x03cf,   -63}, {0x0430, 0x0450,  -32}, {0x0450, 0x0460, -80},
+	{0x0561, 0x0587,   -48}, {0x1f00, 0x1f08,    8}, {0x1f10, 0x1f16,   8},
+	{0x1f20, 0x1f28,     8}, {0x1f30, 0x1f38,    8}, {0x1f40, 0x1f46,   8},
+	{0x1f51, 0x1f52,     8}, {0x1f53, 0x1f54,    8}, {0x1f55, 0x1f56,   8},
+	{0x1f57, 0x1f58,     8}, {0x1f60, 0x1f68,    8}, {0x1f70, 0x1f72,  74},
+	{0x1f72, 0x1f76,    86}, {0x1f76, 0x1f78,  100}, {0x1f78, 0x1f7a, 128},
+	{0x1f7a, 0x1f7c,   112}, {0x1f7c, 0x1f7e,  126}, {0x1f80, 0x1f88,   8},
+	{0x1f90, 0x1f98,     8}, {0x1fa0, 0x1fa8,    8}, {0x1fb0, 0x1fb2,   8},
+	{0x1fb3, 0x1fb4,     9}, {0x1fcc, 0x1fcd,   -9}, {0x1fd0, 0x1fd2,   8},
+	{0x1fe0, 0x1fe2,     8}, {0x1fe5, 0x1fe6,    7}, {0x1ffc, 0x1ffd,  -9},
+	{0x2170, 0x2180,   -16}, {0x24d0, 0x24ea,  -26}, {0x2c30, 0x2c5f, -48},
+	{0x2d00, 0x2d26, -7264}, {0xff41, 0xff5b,  -32}, {0}
 	};
-	static int uc_dup_table[][2] = { /* Start, End */
-	{0x0100, 0x012F}, {0x01A0, 0x01A6}, {0x03E2, 0x03EF}, {0x04CB, 0x04CC},
-	{0x0132, 0x0137}, {0x01B3, 0x01B7}, {0x0460, 0x0481}, {0x04D0, 0x04EB},
-	{0x0139, 0x0149}, {0x01CD, 0x01DD}, {0x0490, 0x04BF}, {0x04EE, 0x04F5},
-	{0x014A, 0x0178}, {0x01DE, 0x01EF}, {0x04BF, 0x04BF}, {0x04F8, 0x04F9},
-	{0x0179, 0x017E}, {0x01F4, 0x01F5}, {0x04C1, 0x04C4}, {0x1E00, 0x1E95},
-	{0x018B, 0x018B}, {0x01FA, 0x0218}, {0x04C7, 0x04C8}, {0x1EA0, 0x1EF9},
-	{0}
+	/*
+	 * "Start" is exclusive and "End" is inclusive, every second value is
+	 * decremented by one.
+	 */
+	static int skip_dec[][2] = { /* Start, End */
+	{0x0100, 0x012f}, {0x0132, 0x0137}, {0x0139, 0x0149}, {0x014a, 0x0178},
+	{0x0179, 0x017e}, {0x01a0, 0x01a6}, {0x01b3, 0x01b7}, {0x01cd, 0x01dd},
+	{0x01de, 0x01ef}, {0x01f4, 0x01f5}, {0x01f8, 0x01f9}, {0x01fa, 0x0220},
+	{0x0222, 0x0234}, {0x023b, 0x023c}, {0x0241, 0x0242}, {0x0246, 0x024f},
+	{0x03d8, 0x03ef}, {0x03f7, 0x03f8}, {0x03fa, 0x03fb}, {0x0460, 0x0481},
+	{0x048a, 0x04bf}, {0x04c1, 0x04c4}, {0x04c5, 0x04c8}, {0x04c9, 0x04ce},
+	{0x04ec, 0x04ed}, {0x04d0, 0x04eb}, {0x04ee, 0x04f5}, {0x04f6, 0x0513},
+	{0x1e00, 0x1e95}, {0x1ea0, 0x1ef9}, {0x2183, 0x2184}, {0x2c60, 0x2c61},
+	{0x2c67, 0x2c6c}, {0x2c75, 0x2c76}, {0x2c80, 0x2ce3}, {0}
 	};
-	static int uc_byte_table[][2] = { /* Offset, Value */
-	{0x00FF, 0x0178}, {0x01AD, 0x01AC}, {0x01F3, 0x01F1}, {0x0269, 0x0196},
-	{0x0183, 0x0182}, {0x01B0, 0x01AF}, {0x0253, 0x0181}, {0x026F, 0x019C},
-	{0x0185, 0x0184}, {0x01B9, 0x01B8}, {0x0254, 0x0186}, {0x0272, 0x019D},
-	{0x0188, 0x0187}, {0x01BD, 0x01BC}, {0x0259, 0x018F}, {0x0275, 0x019F},
-	{0x018C, 0x018B}, {0x01C6, 0x01C4}, {0x025B, 0x0190}, {0x0283, 0x01A9},
-	{0x0192, 0x0191}, {0x01C9, 0x01C7}, {0x0260, 0x0193}, {0x0288, 0x01AE},
-	{0x0199, 0x0198}, {0x01CC, 0x01CA}, {0x0263, 0x0194}, {0x0292, 0x01B7},
-	{0x01A8, 0x01A7}, {0x01DD, 0x018E}, {0x0268, 0x0197},
-	{0}
+	/*
+	 * Set the Unicode character at offset "Offset" to "Value".  Note,
+	 * "Value" is host endian.
+	 */
+	static int set[][2] = { /* Offset, Value */
+	{0x00ff, 0x0178}, {0x0180, 0x0243}, {0x0183, 0x0182}, {0x0185, 0x0184},
+	{0x0188, 0x0187}, {0x018c, 0x018b}, {0x0192, 0x0191}, {0x0195, 0x01f6},
+	{0x0199, 0x0198}, {0x019a, 0x023d}, {0x019e, 0x0220}, {0x01a8, 0x01a7},
+	{0x01ad, 0x01ac}, {0x01b0, 0x01af}, {0x01b9, 0x01b8}, {0x01bd, 0x01bc},
+	{0x01bf, 0x01f7}, {0x01c6, 0x01c4}, {0x01c9, 0x01c7}, {0x01cc, 0x01ca},
+	{0x01dd, 0x018e}, {0x01f3, 0x01f1}, {0x023a, 0x2c65}, {0x023e, 0x2c66},
+	{0x0253, 0x0181}, {0x0254, 0x0186}, {0x0259, 0x018f}, {0x025b, 0x0190},
+	{0x0260, 0x0193}, {0x0263, 0x0194}, {0x0268, 0x0197}, {0x0269, 0x0196},
+	{0x026b, 0x2c62}, {0x026f, 0x019c}, {0x0272, 0x019d}, {0x0275, 0x019f},
+	{0x027d, 0x2c64}, {0x0280, 0x01a6}, {0x0283, 0x01a9}, {0x0288, 0x01ae},
+	{0x0289, 0x0244}, {0x028c, 0x0245}, {0x0292, 0x01b7}, {0x03f2, 0x03f9},
+	{0x04cf, 0x04c0}, {0x1d7d, 0x2c63}, {0x214e, 0x2132}, {0}
 	};
-	int i, r;
+	unsigned i, r;
 
-	memset((char*)uc, 0, uc_len);
-	uc_len >>= 1;
-	if (uc_len > 65536)
-		uc_len = 65536;
-	for (i = 0; (u32)i < uc_len; i++)
-		uc[i] = i;
-	for (r = 0; uc_run_table[r][0]; r++)
-		for (i = uc_run_table[r][0]; i < uc_run_table[r][1]; i++)
-			uc[i] += uc_run_table[r][2];
-	for (r = 0; uc_dup_table[r][0]; r++)
-		for (i = uc_dup_table[r][0]; i < uc_dup_table[r][1]; i += 2)
-			uc[i + 1]--;
-	for (r = 0; uc_byte_table[r][0]; r++)
-		uc[uc_byte_table[r][0]] = uc_byte_table[r][1];
+	memset(uc, 0, uc_len);
+	uc_len /= 2;
+	/* Start with a one-to-one mapping, i.e. no upcasing happens at all. */
+	for (i = 0; i < uc_len; i++)
+		uc[i] = cpu_to_le16(i);
+	/* Adjust specified runs by the specified amount. */
+	for (r = 0; add[r][0]; r++)
+		for (i = add[r][0]; i < add[r][1]; i++)
+			uc[i] = cpu_to_le16(le16_to_cpu(uc[i]) + add[r][2]);
+	/* Decrement every second value in specified runs. */
+	for (r = 0; skip_dec[r][0]; r++)
+		for (i = skip_dec[r][0]; i < skip_dec[r][1];
+				i += 2)
+			uc[i + 1] = cpu_to_le16(le16_to_cpu(uc[i + 1]) - 1);
+	/* Set specified characters to specified values. */
+	for (r = 0; set[r][0]; r++)
+		uc[set[r][0]] = cpu_to_le16(set[r][1]);
 }
+
+/**
+ * ntfs_str2ucs - convert a string to a valid NTFS file name
+ * @s:		input string
+ * @len:	length of output buffer in Unicode characters
+ *
+ * Convert the input @s string into the corresponding little endian,
+ * 2-byte Unicode string. The length of the converted string is less
+ * or equal to the maximum length allowed by the NTFS format (255).
+ *
+ * If @s is NULL then return AT_UNNAMED.
+ *
+ * On success the function returns the Unicode string in an allocated
+ * buffer and the caller is responsible to free it when it's not needed
+ * anymore.
+ *
+ * On error NULL is returned and errno is set to the error code.
+ */
+ntfschar *ntfs_str2ucs(const char *s, int *len)
+{
+	ntfschar *ucs = NULL;
+
+	if (s && ((*len = ntfs_mbstoucs(s, &ucs, 0)) == -1)) {
+		ntfs_log_perror("Couldn't convert '%s' to Unicode", s);
+		return NULL;
+	}
+	if (*len > NTFS_MAX_NAME_LEN) {
+		free(ucs);
+		errno = ENAMETOOLONG;
+		return NULL;
+	}
+	if (!ucs || !*len) {
+		ucs  = AT_UNNAMED;
+		*len = 0;
+	}
+	return ucs;
+}
+
+/**
+ * ntfs_ucsfree - free memory allocated by ntfs_str2ucs()
+ * @ucs:	input string to be freed
+ *
+ * Free memory at @ucs and which was allocated by ntfs_str2ucs.
+ *
+ * Return value: none.
+ */
+void ntfs_ucsfree(ntfschar *ucs)
+{
+	if (ucs && (ucs != AT_UNNAMED))
+		free(ucs);
+}
+
