@@ -1,7 +1,7 @@
-/*
+/**
  * unix_io.c - Unix style disk io functions. Part of the Linux-NTFS project.
  *
- * Copyright (c) 2000-2003 Anton Altaparmakov
+ * Copyright (c) 2000-2006 Anton Altaparmakov
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -19,7 +19,9 @@
  * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -49,33 +51,45 @@
 #include <sys/ioctl.h>
 #endif
 #ifdef HAVE_LINUX_FD_H
-#	include <linux/fd.h>
+#include <linux/fd.h>
 #endif
 
 #include "types.h"
 #include "mst.h"
 #include "debug.h"
 #include "device.h"
-
-#if defined(linux) && defined(_IO) && !defined(BLKGETSIZE)
-#	define BLKGETSIZE _IO(0x12,96) /* Get device size in 512byte blocks. */
-#endif
+#include "logging.h"
 
 #define DEV_FD(dev)	(*(int *)dev->d_private)
 
+/* Define to nothing if not present on this system. */
+#ifndef O_EXCL
+#	define O_EXCL 0
+#endif
+
+/**
+ * ntfs_device_unix_io_open - Open a device and lock it exclusively
+ * @dev:
+ * @flags:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static int ntfs_device_unix_io_open(struct ntfs_device *dev, int flags)
 {
 	struct flock flk;
+	struct stat sbuf;
 	int err;
 
 	if (NDevOpen(dev)) {
 		errno = EBUSY;
 		return -1;
 	}
-	if (!(dev->d_private = malloc(sizeof(int))))
+	if (!(dev->d_private = ntfs_malloc(sizeof(int))))
 		return -1;
-	/* Open the device/file obtaining the file descriptor. */
-	if ((*(int *)dev->d_private = open(dev->d_name, flags)) == -1) {
+	*(int*)dev->d_private = open(dev->d_name, flags);
+	if (*(int*)dev->d_private == -1) {
 		err = errno;
 		goto err_out;
 	}
@@ -92,15 +106,17 @@ static int ntfs_device_unix_io_open(struct ntfs_device *dev, int flags)
 	flk.l_start = flk.l_len = 0LL;
 	if (fcntl(DEV_FD(dev), F_SETLK, &flk)) {
 		err = errno;
-		Dprintf("ntfs_device_unix_io_open: Could not lock %s for %s: "
-				"%s\n", dev->d_name, NDevReadOnly(dev) ?
-				"reading" : "writing", strerror(errno));
+		ntfs_log_debug("ntfs_device_unix_io_open: Could not lock %s "
+				"for %s\n", dev->d_name, NDevReadOnly(dev) ?
+				"reading" : "writing");
 		if (close(DEV_FD(dev)))
-			Dprintf("ntfs_device_unix_io_open: Warning: Could not "
-					"close %s: %s\n", dev->d_name,
-					strerror(errno));
+			ntfs_log_perror("ntfs_device_unix_io_open: Warning: "
+					"Could not close %s", dev->d_name);
 		goto err_out;
 	}
+	/* Determine if device is a block device or not, ignoring errors. */
+	if (!fstat(DEV_FD(dev), &sbuf) && S_ISBLK(sbuf.st_mode))
+		NDevSetBlock(dev);
 	/* Set our open flag. */
 	NDevSetOpen(dev);
 	return 0;
@@ -111,6 +127,14 @@ err_out:
 	return -1;
 }
 
+/**
+ * ntfs_device_unix_io_close - Close the device, releasing the lock
+ * @dev:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static int ntfs_device_unix_io_close(struct ntfs_device *dev)
 {
 	struct flock flk;
@@ -127,8 +151,8 @@ static int ntfs_device_unix_io_close(struct ntfs_device *dev)
 	flk.l_whence = SEEK_SET;
 	flk.l_start = flk.l_len = 0LL;
 	if (fcntl(DEV_FD(dev), F_SETLK, &flk))
-		Dprintf("ntfs_device_unix_io_close: Warning: Could not unlock "
-				"%s: %s\n", dev->d_name, strerror(errno));
+		ntfs_log_perror("ntfs_device_unix_io_close: Warning: Could not "
+				"unlock %s", dev->d_name);
 	/* Close the file descriptor and clear our open flag. */
 	if (close(DEV_FD(dev)))
 		return -1;
@@ -138,18 +162,48 @@ static int ntfs_device_unix_io_close(struct ntfs_device *dev)
 	return 0;
 }
 
+/**
+ * ntfs_device_unix_io_seek - Seek to a place on the device
+ * @dev:
+ * @offset:
+ * @whence:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static s64 ntfs_device_unix_io_seek(struct ntfs_device *dev, s64 offset,
 		int whence)
 {
 	return lseek(DEV_FD(dev), offset, whence);
 }
 
+/**
+ * ntfs_device_unix_io_read - Read from the device, from the current location
+ * @dev:
+ * @buf:
+ * @count:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static s64 ntfs_device_unix_io_read(struct ntfs_device *dev, void *buf,
 		s64 count)
 {
 	return read(DEV_FD(dev), buf, count);
 }
 
+/**
+ * ntfs_device_unix_io_write - Write to the device, at the current location
+ * @dev:
+ * @buf:
+ * @count:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static s64 ntfs_device_unix_io_write(struct ntfs_device *dev, const void *buf,
 		s64 count)
 {
@@ -161,12 +215,34 @@ static s64 ntfs_device_unix_io_write(struct ntfs_device *dev, const void *buf,
 	return write(DEV_FD(dev), buf, count);
 }
 
+/**
+ * ntfs_device_unix_io_pread - Perform a positioned read from the device
+ * @dev:
+ * @buf:
+ * @count:
+ * @offset:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static s64 ntfs_device_unix_io_pread(struct ntfs_device *dev, void *buf,
 		s64 count, s64 offset)
 {
-	return ntfs_pread(dev, offset, count, buf);
+	return pread(DEV_FD(dev), buf, count, offset);
 }
 
+/**
+ * ntfs_device_unix_io_pwrite - Perform a positioned write to the device
+ * @dev:
+ * @buf:
+ * @count:
+ * @offset:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static s64 ntfs_device_unix_io_pwrite(struct ntfs_device *dev, const void *buf,
 		s64 count, s64 offset)
 {
@@ -175,9 +251,17 @@ static s64 ntfs_device_unix_io_pwrite(struct ntfs_device *dev, const void *buf,
 		return -1;
 	}
 	NDevSetDirty(dev);
-	return ntfs_pwrite(dev, offset, count, buf);
+	return pwrite(DEV_FD(dev), buf, count, offset);
 }
 
+/**
+ * ntfs_device_unix_io_sync - Flush any buffered changes to the device
+ * @dev:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static int ntfs_device_unix_io_sync(struct ntfs_device *dev)
 {
 	if (!NDevReadOnly(dev) && NDevDirty(dev)) {
@@ -189,11 +273,30 @@ static int ntfs_device_unix_io_sync(struct ntfs_device *dev)
 	return 0;
 }
 
+/**
+ * ntfs_device_unix_io_stat - Get information about the device
+ * @dev:
+ * @buf:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static int ntfs_device_unix_io_stat(struct ntfs_device *dev, struct stat *buf)
 {
 	return fstat(DEV_FD(dev), buf);
 }
 
+/**
+ * ntfs_device_unix_io_ioctl - Perform an ioctl on the device
+ * @dev:
+ * @request:
+ * @argp:
+ *
+ * Description...
+ *
+ * Returns:
+ */
 static int ntfs_device_unix_io_ioctl(struct ntfs_device *dev, int request,
 		void *argp)
 {
@@ -215,4 +318,3 @@ struct ntfs_device_operations ntfs_device_unix_io_ops = {
 	.stat		= ntfs_device_unix_io_stat,
 	.ioctl		= ntfs_device_unix_io_ioctl,
 };
-
