@@ -1,9 +1,10 @@
 /**
  * ntfsresize - Part of the Linux-NTFS project.
  *
- * Copyright (c) 2002-2005 Szabolcs Szakacsits
+ * Copyright (c) 2002-2006 Szabolcs Szakacsits
  * Copyright (c) 2002-2005 Anton Altaparmakov
  * Copyright (c) 2002-2003 Richard Russon
+ * Copyright (c) 2007      Yura Pakhuchiy
  *
  * This utility will resize an NTFS volume without data loss.
  *
@@ -91,9 +92,9 @@ static const char *invalid_ntfs_msg =
 "if the disk was incorrectly repartitioned (see the ntfsresize FAQ).\n";
 
 static const char *corrupt_volume_msg =
-"This software has detected that your NTFS is corrupted. Please run chkdsk /f\n"
-"on Windows then reboot it TWICE! Important, don't forget the /f parameter!\n"
-"Afterwards you can run ntfsresize. No modification was made to NTFS.\n";
+"NTFS is inconsistent. Run chkdsk /f on Windows then reboot it TWICE!\n"
+"The usage of the /f parameter is very IMPORTANT! No modification was\n"
+"and will be made to NTFS by this software until it gets repaired.\n";
 
 static const char *hibernated_volume_msg =
 "The NTFS partition is hibernated. Windows must be resumed and turned off\n"
@@ -104,6 +105,10 @@ static const char *unclean_journal_msg =
 "using this software! Note, if you have run chkdsk previously then boot\n"
 "Windows again which will automatically initialize the journal correctly.\n";
 
+static const char *opened_volume_msg =
+"This software has detected that the NTFS volume is already opened by another\n"
+"software thus it refuses to progress to preserve data consistency.\n";
+
 static const char *bad_sectors_warning_msg =
 "****************************************************************************\n"
 "* WARNING: The disk has bad sector. This means physical damage on the disk *\n"
@@ -114,7 +119,14 @@ static const char *bad_sectors_warning_msg =
 "* NTFS safely by additionally using the --bad-sectors option of ntfsresize.*\n"
 "****************************************************************************\n";
 
-struct {
+static const char *many_bad_sectors_msg =
+"***************************************************************************\n"
+"* WARNING: The disk has many bad sectors. This means physical damage      *\n"
+"* on the disk surface caused by deterioration, manufacturing faults or    *\n"
+"* other reason. We suggest to get a replacement disk as soon as possible. *\n"
+"***************************************************************************\n";
+
+static struct {
 	int verbose;
 	int debug;
 	int ro_flag;
@@ -129,7 +141,6 @@ struct {
 struct bitmap {
 	s64 size;
 	u8 *bm;
-	u8 padding[4];		/* Unused: padding to 64 bit. */
 };
 
 #define NTFS_PROGBAR		0x0001
@@ -141,7 +152,6 @@ struct progress_bar {
 	int resolution;
 	int flags;
 	float unit;
-	u8 padding[4];		/* Unused: padding to 64 bit. */
 };
 
 struct llcn_t {
@@ -191,7 +201,7 @@ typedef struct {
 
 /* FIXME: This, lcn_bitmap and pos from find_free_cluster() will make a cluster
    allocation related structure, attached to ntfs_resize_t */
-s64 max_free_cluster_range = 0;
+static s64 max_free_cluster_range = 0;
 
 #define NTFS_MBYTE (1000 * 1000)
 
@@ -204,11 +214,7 @@ s64 max_free_cluster_range = 0;
 #define DIRTY_INODE		(1)
 #define DIRTY_ATTRIB		(2)
 
-#define NTFS_MAX_CLUSTER_SIZE 	(65536)
-
-GEN_PRINTF(Eprintf, stderr, NULL,         FALSE)
-GEN_PRINTF(Vprintf, stdout, &opt.verbose, TRUE)
-GEN_PRINTF(Qprintf, stdout, NULL,         FALSE)
+#define NTFS_MAX_CLUSTER_SIZE	(65536)
 
 static s64 rounded_up_division(s64 numer, s64 denom)
 {
@@ -220,8 +226,7 @@ static s64 rounded_up_division(s64 numer, s64 denom)
  *
  * Print an error message.
  */
-static void perr_printf(const char *fmt, ...)
-		__attribute__((format(printf, 1, 2)));
+__attribute__((format(printf, 1, 2)))
 static void perr_printf(const char *fmt, ...)
 {
 	va_list ap;
@@ -236,8 +241,7 @@ static void perr_printf(const char *fmt, ...)
 	fflush(stderr);
 }
 
-static void err_printf(const char *fmt, ...)
-		__attribute__((format(printf, 1, 2)));
+__attribute__((format(printf, 1, 2)))
 static void err_printf(const char *fmt, ...)
 {
 	va_list ap;
@@ -255,9 +259,8 @@ static void err_printf(const char *fmt, ...)
  *
  * Print and error message and exit the program.
  */
-static int err_exit(const char *fmt, ...)
-		__attribute__((noreturn))
-		__attribute__((format(printf, 1, 2)));
+__attribute__((noreturn))
+__attribute__((format(printf, 1, 2)))
 static int err_exit(const char *fmt, ...)
 {
 	va_list ap;
@@ -276,9 +279,8 @@ static int err_exit(const char *fmt, ...)
  *
  * Print and error message and exit the program
  */
-static int perr_exit(const char *fmt, ...)
-		__attribute__((noreturn))
-		__attribute__((format(printf, 1, 2)));
+__attribute__((noreturn))
+__attribute__((format(printf, 1, 2)))
 static int perr_exit(const char *fmt, ...)
 {
 	va_list ap;
@@ -301,11 +303,11 @@ static int perr_exit(const char *fmt, ...)
  *
  * Return:  none
  */
-static void usage(void) __attribute__((noreturn));
+__attribute__((noreturn))
 static void usage(void)
 {
 
-	printf ("\nUsage: %s [OPTIONS] DEVICE\n"
+	printf("\nUsage: %s [OPTIONS] DEVICE\n"
 		"    Resize an NTFS volume non-destructively, safely move any data if needed.\n"
 		"\n"
 		"    -i, --info             Estimate the smallest shrunken size possible\n"
@@ -325,8 +327,8 @@ static void usage(void)
 		"    The options -i and -s are mutually exclusive. If both options are\n"
 		"    omitted then the NTFS volume will be enlarged to the DEVICE size.\n"
 		"\n", EXEC_NAME);
-	printf ("%s%s", ntfs_bugs, ntfs_home);
-	printf ("Ntfsresize FAQ: http://linux-ntfs.sourceforge.net/info/ntfsresize.html\n");
+	printf("%s%s", ntfs_bugs, ntfs_home);
+	printf("Ntfsresize FAQ: http://linux-ntfs.sourceforge.net/info/ntfsresize.html\n");
 	exit(1);
 }
 
@@ -346,7 +348,7 @@ static void proceed_question(void)
 	printf("Are you sure you want to proceed (y/[n])? ");
 	buf[0] = 0;
 	fgets(buf, sizeof(buf), stdin);
-	if (strchr(short_yes, buf[0]) == 0) {
+	if (!strchr(short_yes, buf[0])) {
 		printf("OK quitting. NO CHANGES have been made to your "
 				"NTFS volume.\n");
 		exit(1);
@@ -360,13 +362,14 @@ static void proceed_question(void)
  *
  * Return:  none
  */
-static void version (void)
+static void version(void)
 {
-	printf ("\nResize an NTFS Volume, without data loss.\n\n");
-	printf ("Copyright (c) 2002-2005  Szabolcs Szakacsits\n");
-	printf ("Copyright (c) 2002-2004  Anton Altaparmakov\n");
-	printf ("Copyright (c) 2002-2003  Richard Russon\n");
-	printf ("\n%s\n%s%s", ntfs_gpl, ntfs_bugs, ntfs_home);
+	printf("\nResize an NTFS Volume, without data loss.\n\n");
+	printf("Copyright (c) 2002-2006  Szabolcs Szakacsits\n");
+	printf("Copyright (c) 2002-2005  Anton Altaparmakov\n");
+	printf("Copyright (c) 2002-2003  Richard Russon\n");
+	printf("Copyright (c) 2007       Yura Pakhuchiy\n");
+	printf("\n%s\n%s%s", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
 
 /**
@@ -450,7 +453,7 @@ static int parse_options(int argc, char **argv)
 		{ NULL, 0, NULL, 0 }
 	};
 
-	char c;
+	int c;
 	int err  = 0;
 	int ver  = 0;
 	int help = 0;
@@ -458,7 +461,7 @@ static int parse_options(int argc, char **argv)
 	memset(&opt, 0, sizeof(opt));
 	opt.show_progress = 1;
 
-	while ((c = getopt_long (argc, argv, sopt, lopt, NULL)) != (char)-1) {
+	while ((c = getopt_long(argc, argv, sopt, lopt, NULL)) != -1) {
 		switch (c) {
 		case 1:	/* A non-option argument */
 			if (!err && !opt.volume)
@@ -483,7 +486,7 @@ static int parse_options(int argc, char **argv)
 			opt.info++;
 			break;
 		case 'n':
-			opt.ro_flag = MS_RDONLY;
+			opt.ro_flag = NTFS_MNT_RDONLY;
 			break;
 		case 'P':
 			opt.show_progress = 0;
@@ -496,15 +499,16 @@ static int parse_options(int argc, char **argv)
 			break;
 		case 'v':
 			opt.verbose++;
+			ntfs_log_set_levels(NTFS_LOG_LEVEL_VERBOSE);
 			break;
 		case 'V':
 			ver++;
 			break;
 		default:
 			if (optopt == 's') {
-				Eprintf("Option '%s' requires an argument.\n", argv[optind-1]);
+				printf("Option '%s' requires an argument.\n", argv[optind-1]);
 			} else {
-				Eprintf("Unknown option '%s'.\n", argv[optind-1]);
+				printf("Unknown option '%s'.\n", argv[optind-1]);
 			}
 			err++;
 			break;
@@ -514,25 +518,31 @@ static int parse_options(int argc, char **argv)
 	if (!help && !ver) {
 		if (opt.volume == NULL) {
 			if (argc > 1)
-				Eprintf("You must specify exactly one device.\n");
+				printf("You must specify exactly one device.\n");
 			err++;
 		}
 		if (opt.info) {
-			opt.ro_flag = MS_RDONLY;
+			opt.ro_flag = NTFS_MNT_RDONLY;
 			if (opt.bytes) {
-				Eprintf(NERR_PREFIX "Options --info and --size "
+				printf(NERR_PREFIX "Options --info and --size "
 					"can't be used together.\n");
 				usage();
 			}
 		}
 	}
 
-	stderr = stdout;
+	/* Redirect stderr to stdout, note fflush()es are essential! */
+	fflush(stdout);
+	fflush(stderr);
+	if (dup2(STDOUT_FILENO, STDERR_FILENO) == -1)
+		perr_exit("Failed to redirect stderr to stdout");
+	fflush(stdout);
+	fflush(stderr);
 
 #ifdef DEBUG
 	if (!opt.debug)
-		if (!(stderr = fopen("/dev/null", "rw")))
-			perr_exit("Couldn't open /dev/null");
+		if (!freopen("/dev/null", "w", stderr))
+			perr_exit("Failed to redirect stderr to /dev/null");
 #endif
 
 	if (ver)
@@ -601,9 +611,9 @@ static int rl_items(runlist *rl)
 
 static void dump_run(runlist_element *r)
 {
-	Vprintf(" %8lld  %8lld (0x%08llx)  %lld\n", (long long)r->vcn,
-			(long long)r->lcn, (long long)r->lcn,
-			(long long)r->length);
+	ntfs_log_verbose(" %8lld  %8lld (0x%08llx)  %lld\n", (long long)r->vcn,
+			 (long long)r->lcn, (long long)r->lcn,
+			 (long long)r->length);
 }
 
 static void dump_runlist(runlist *rl)
@@ -628,49 +638,6 @@ static s64 nr_clusters_to_bitmap_byte_size(s64 nr_clusters)
 	return bm_bsize;
 }
 
-static int str2unicode(const char *aname, ntfschar **ustr, int *len)
-{
-	if (aname && ((*len = ntfs_mbstoucs(aname, ustr, 0)) == -1)) {
-		perr_printf("Couldn't convert string to Unicode");
-		return -1;
-	}
-
-	if (!*ustr || !*len) {
-		*ustr = AT_UNNAMED;
-		*len = 0;
-	}
-
-	return 0;
-}
-
-static int has_bad_sectors(ntfs_resize_t *resize, int is_inode)
-{
-	int len, ret = 0;
-	ntfschar *ustr = NULL;
-	ATTR_RECORD *a = resize->ctx->attr;
-
-	if (is_inode) {
-		if (resize->ni->mft_no != FILE_BadClus)
-			return 0;
-	} else {
-		if (resize->mref != FILE_BadClus)
-			return 0;
-	}
-
-	if (str2unicode("$Bad", &ustr, &len) == -1)
-		return -1;
-
-	if (ustr && ntfs_names_are_equal(ustr, len,
-			(ntfschar*)((u8*)a + le16_to_cpu(a->name_offset)),
-			a->name_length, 0, NULL, 0))
-		ret = 1;
-
-	if (ustr != AT_UNNAMED)
-		free(ustr);
-
-	return ret;
-}
-
 static void collect_resize_constraints(ntfs_resize_t *resize, runlist *rl)
 {
 	s64 inode, last_lcn;
@@ -685,19 +652,12 @@ static void collect_resize_constraints(ntfs_resize_t *resize, runlist *rl)
 	flags = resize->ctx->attr->flags;
 	atype = resize->ctx->attr->type;
 
-	if ((ret = has_bad_sectors(resize, 1)) != 0) {
+	if ((ret = ntfs_inode_badclus_bad(inode, resize->ctx->attr)) != 0) {
 		if (ret == -1)
-			exit(1);
-		if (NInoAttrList(resize->ni))
-			err_exit("Hopelessly many bad sectors! Not supported.");
-		if (resize->badclusters == 0)
-			printf("WARNING: The disk has bad sector! "
-			       "This can cause reliability problems!\n");
-		resize->badclusters += last_lcn - rl->lcn + 1;
-		Vprintf("Bad clusters: %8lld - %lld\n", rl->lcn, last_lcn);
+			perr_exit("Bad sector list check failed");
 		return;
 	}
-	
+
 	if (inode == FILE_Bitmap) {
 		llcn = &resize->last_lcn;
 		if (atype == AT_DATA && NInoAttrList(resize->ni))
@@ -768,7 +728,7 @@ static void collect_relocation_info(ntfs_resize_t *resize, runlist *rl)
 
 	if (inode == FILE_Bitmap && resize->ctx->attr->type == AT_DATA)
 		return;
-	
+
 	start = lcn;
 	len = lcn_length;
 
@@ -843,11 +803,9 @@ static void build_lcn_usage_bitmap(ntfs_volume *vol, ntfsck_t *fsck)
 				 lcn_length);
 
 		for (j = 0; j < lcn_length; j++) {
-
 			u64 k = (u64)lcn + j;
 
 			if (k >= (u64)vol->nr_clusters) {
-
 				long long outsiders = lcn_length - j;
 
 				fsck->outsider += outsiders;
@@ -863,7 +821,7 @@ static void build_lcn_usage_bitmap(ntfs_volume *vol, ntfsck_t *fsck)
 			if (ntfs_bit_get_and_set(lcn_bitmap->bm, k, 1)) {
 				if (++fsck->multi_ref <= 10 || opt.verbose)
 					printf("Cluster %lld is referenced "
-					       "multiply times!\n",
+					       "multiple times!\n",
 					       (long long)k);
 				continue;
 			}
@@ -958,7 +916,7 @@ static void compare_bitmaps(ntfs_volume *vol, struct bitmap *a)
 					continue;
 				}
 
-				if (++mismatch > 10)
+				if (++mismatch > 10 && !opt.verbose)
 					continue;
 
 				printf("Cluster accounting failed at %lld "
@@ -971,9 +929,9 @@ static void compare_bitmaps(ntfs_volume *vol, struct bitmap *a)
 	}
 done:
 	if (mismatch) {
-		err_printf("Filesystem check failed! Totally %d cluster "
-			   "accounting mismatches.\n", mismatch);
-		printf("%s", corrupt_volume_msg);
+		printf("Filesystem check failed! Totally %d cluster "
+		       "accounting mismatches.\n", mismatch);
+		err_printf("%s", corrupt_volume_msg);
 		exit(1);
 	}
 }
@@ -1153,7 +1111,7 @@ static void rl_fixup(runlist **rl)
 	if (tmp->lcn == LCN_RL_NOT_MAPPED) {
 		s64 unmapped_len = tmp->length;
 
-		Vprintf("Skip unmapped run at the beginning ...\n");
+		ntfs_log_verbose("Skip unmapped run at the beginning ...\n");
 
 		if (!tmp->length)
 			err_exit("Empty unmapped runlist! Please report!\n");
@@ -1164,7 +1122,7 @@ static void rl_fixup(runlist **rl)
 
 	for (tmp = *rl; tmp->length; tmp++) {
 		if (tmp->lcn == LCN_RL_NOT_MAPPED) {
-			Vprintf("Skip unmapped run at the end  ...\n");
+			ntfs_log_verbose("Skip unmapped run at the end  ...\n");
 
 			if (tmp[1].length)
 				err_exit("Unmapped runlist in the middle! "
@@ -1204,28 +1162,29 @@ static void replace_attribute_runlist(ntfs_volume *vol,
 		s64 remains_size;
 		char *next_attr;
 
-		Vprintf("Enlarging attribute header ...\n");
+		ntfs_log_verbose("Enlarging attribute header ...\n");
 
 		mp_size = (mp_size + 7) & ~7;
 
-		Vprintf("Old mp size      : %d\n", l);
-		Vprintf("New mp size      : %d\n", mp_size);
-		Vprintf("Bytes in use     : %u\n", (unsigned int)
-				le32_to_cpu(ctx->mrec->bytes_in_use));
+		ntfs_log_verbose("Old mp size      : %d\n", l);
+		ntfs_log_verbose("New mp size      : %d\n", mp_size);
+		ntfs_log_verbose("Bytes in use     : %u\n", (unsigned int)
+				 le32_to_cpu(ctx->mrec->bytes_in_use));
 
-		next_attr = (char *)a + le16_to_cpu(a->length);
+		next_attr = (char *)a + le32_to_cpu(a->length);
 		l = mp_size - l;
 
-		Vprintf("Bytes in use new : %u\n", l + (unsigned int)
-				le32_to_cpu(ctx->mrec->bytes_in_use));
-		Vprintf("Bytes allocated  : %u\n", (unsigned int)
-				le32_to_cpu(ctx->mrec->bytes_allocated));
+		ntfs_log_verbose("Bytes in use new : %u\n", l + (unsigned int)
+				 le32_to_cpu(ctx->mrec->bytes_in_use));
+		ntfs_log_verbose("Bytes allocated  : %u\n", (unsigned int)
+				 le32_to_cpu(ctx->mrec->bytes_allocated));
 
 		remains_size = le32_to_cpu(ctx->mrec->bytes_in_use);
 		remains_size -= (next_attr - (char *)ctx->mrec);
 
-		Vprintf("increase         : %d\n", l);
-		Vprintf("shift            : %lld\n", (long long)remains_size);
+		ntfs_log_verbose("increase         : %d\n", l);
+		ntfs_log_verbose("shift            : %lld\n",
+				 (long long)remains_size);
 
 		if (le32_to_cpu(ctx->mrec->bytes_in_use) + l >
 				le32_to_cpu(ctx->mrec->bytes_allocated))
@@ -1239,11 +1198,12 @@ static void replace_attribute_runlist(ntfs_volume *vol,
 		memmove(next_attr + l, next_attr, remains_size);
 		ctx->mrec->bytes_in_use = cpu_to_le32(l +
 				le32_to_cpu(ctx->mrec->bytes_in_use));
-		a->length += l;
+		a->length = cpu_to_le32(le32_to_cpu(a->length) + l);
 	}
 
-	if (!(mp = calloc(1, mp_size)))
-		perr_exit("Couldn't get memory");
+	mp = ntfs_calloc(mp_size);
+	if (!mp)
+		perr_exit("ntfsc_calloc couldn't get memory");
 
 	if (ntfs_mapping_pairs_build(vol, mp, mp_size, rl, 0, NULL))
 		perr_exit("ntfs_mapping_pairs_build");
@@ -1325,8 +1285,8 @@ static int find_free_cluster(struct bitmap *bm,
 	}
 	if (rle->length < items && rle->length < max_free_cluster_range) {
 		max_free_cluster_range = rle->length;
-		Vprintf("Max free range: %7lld     \n",
-				(long long)max_free_cluster_range);
+		ntfs_log_verbose("Max free range: %7lld     \n",
+				 (long long)max_free_cluster_range);
 	}
 	pos = rle->lcn + items;
 	if (pos == nr_vol_clusters)
@@ -1373,7 +1333,7 @@ static runlist *alloc_cluster(struct bitmap *bm,
 	rl_set(rl + runs, vcn, -1LL, 0LL);
 
 	if (runs > 1) {
-		Vprintf("Multi-run allocation:    \n");
+		ntfs_log_verbose("Multi-run allocation:    \n");
 		dump_runlist(rl);
 	}
 	return rl;
@@ -1459,13 +1419,21 @@ static void copy_clusters(ntfs_resize_t *resize, s64 dest, s64 src, s64 len)
 
 		lseek_to_cluster(vol, src + i);
 
-		if (read_all(vol->dev, buff, vol->cluster_size) == -1)
-			perr_exit("read_all");
+		if (read_all(vol->dev, buff, vol->cluster_size) == -1) {
+			perr_printf("Failed to read from the disk");
+			if (errno == EIO)
+				printf("%s", bad_sectors_warning_msg);
+			exit(1);
+		}
 
 		lseek_to_cluster(vol, dest + i);
 
-		if (write_all(vol->dev, buff, vol->cluster_size) == -1)
-			perr_exit("write_all");
+		if (write_all(vol->dev, buff, vol->cluster_size) == -1) {
+			perr_printf("Failed to write to the disk");
+			if (errno == EIO)
+				printf("%s", bad_sectors_warning_msg);
+			exit(1);
+		}
 
 		resize->relocations++;
 		progress_update(&resize->progress, resize->relocations);
@@ -1499,8 +1467,9 @@ static void rl_split_run(runlist **rl, int run, s64 pos)
 	size_head = run * sizeof(runlist_element);
 	size_tail = (items - run - 1) * sizeof(runlist_element);
 
-	if (!(rl_new = (runlist *)malloc(new_size)))
-		perr_exit("malloc");
+	rl_new = ntfs_malloc(new_size);
+	if (!rl_new)
+		perr_exit("ntfs_malloc");
 
 	rle_new = rl_new + run;
 	rle = *rl + run;
@@ -1514,7 +1483,7 @@ static void rl_split_run(runlist **rl, int run, s64 pos)
 	rl_set(rle_new, rle->vcn, rle->lcn, len_head);
 	rl_set(rle_new + 1, rle->vcn + len_head, rle->lcn + len_head, len_tail);
 
-	Vprintf("Splitting run at cluster %lld:\n", (long long)pos);
+	ntfs_log_verbose("Splitting run at cluster %lld:\n", (long long)pos);
 	dump_run(rle); dump_run(rle_new); dump_run(rle_new + 1);
 
 	free(*rl);
@@ -1577,11 +1546,13 @@ static void relocate_run(ntfs_resize_t *resize, runlist **rl, int run)
 			  resize->mref, lcn_length);
 
 	/* FIXME: check $MFTMirr DATA isn't multi-run (or support it) */
-	Vprintf("Relocate inode %7llu:0x%x:%08lld:0x%08llx --> 0x%08llx\n",
-			(unsigned long long)resize->mref,
-			(unsigned int)le32_to_cpu(resize->ctx->attr->type),
-			(long long)lcn_length, (unsigned long long)lcn,
-			(unsigned long long)relocate_rl->lcn);
+	ntfs_log_verbose("Relocate record %7llu:0x%x:%08lld:0x%08llx:0x%08llx "
+			 "--> 0x%08llx\n", (unsigned long long)resize->mref,
+			 (unsigned int)le32_to_cpu(resize->ctx->attr->type),
+			 (long long)lcn_length,
+			 (unsigned long long)(*rl + run)->vcn,
+			 (unsigned long long)lcn,
+			 (unsigned long long)relocate_rl->lcn);
 
 	relocate_clusters(resize, relocate_rl, lcn);
 	rl_insert_at_run(rl, run, relocate_rl);
@@ -1639,11 +1610,11 @@ static int is_mftdata(ntfs_resize_t *resize)
 
 	if (resize->mref == 0)
 		return 1;
-	
-	if (  MREF(resize->mrec->base_mft_record) == 0  &&
-	    MSEQNO(resize->mrec->base_mft_record) != 0)
+
+	if (MREF_LE(resize->mrec->base_mft_record) == 0 &&
+	    MSEQNO_LE(resize->mrec->base_mft_record) != 0)
 		return 1;
-	
+
 	return 0;
 }
 
@@ -1651,30 +1622,30 @@ static int handle_mftdata(ntfs_resize_t *resize, int do_mftdata)
 {
 	ATTR_RECORD *attr = resize->ctx->attr;
 	VCN highest_vcn, lowest_vcn;
-	
+
 	if (do_mftdata) {
-		
+
 		if (!is_mftdata(resize))
 			return 0;
-		
+
 		highest_vcn = sle64_to_cpu(attr->highest_vcn);
 		lowest_vcn  = sle64_to_cpu(attr->lowest_vcn);
-		
+
 		if (resize->mft_highest_vcn != highest_vcn)
 			return 0;
-		
+
 		if (lowest_vcn == 0)
 			resize->mft_highest_vcn = lowest_vcn;
 		else
 			resize->mft_highest_vcn = lowest_vcn - 1;
 
 	} else if (is_mftdata(resize)) {
-		
+
 		highest_vcn = sle64_to_cpu(attr->highest_vcn);
-		
+
 		if (resize->mft_highest_vcn < highest_vcn)
 			resize->mft_highest_vcn = highest_vcn;
-		
+
 		return 0;
 	}
 
@@ -1691,13 +1662,13 @@ static void relocate_attributes(ntfs_resize_t *resize, int do_mftdata)
 	while (!ntfs_attrs_walk(resize->ctx)) {
 		if (resize->ctx->attr->type == AT_END)
 			break;
-		
+
 		if (handle_mftdata(resize, do_mftdata) == 0)
 			continue;
-			
-		ret = has_bad_sectors(resize, 0);
+
+		ret = ntfs_inode_badclus_bad(resize->mref, resize->ctx->attr);
 		if (ret == -1)
-			exit(1);
+			perr_exit("Bad sector list check failed");
 		else if (ret == 1)
 			continue;
 
@@ -1733,7 +1704,7 @@ static void relocate_inode(ntfs_resize_t *resize, MFT_REF mref, int do_mftdata)
 //		if (vol->dev->d_ops->sync(vol->dev) == -1)
 //			perr_exit("Failed to sync device");
 		if (write_mft_record(resize->vol, mref, resize->mrec))
-			perr_exit("Couldn't update inode %llu", mref);
+			perr_exit("Couldn't update record %llu", mref);
 	}
 }
 
@@ -1748,17 +1719,17 @@ static void relocate_inodes(ntfs_resize_t *resize)
 	progress_init(&resize->progress, 0, resize->relocations, resize->progress.flags);
 	resize->relocations = 0;
 
-	resize->mrec = (MFT_RECORD *)malloc(resize->vol->mft_record_size);
+	resize->mrec = ntfs_malloc(resize->vol->mft_record_size);
 	if (!resize->mrec)
-		perr_exit("malloc failed");
+		perr_exit("ntfs_malloc failed");
 
 	nr_mft_records = resize->vol->mft_na->initialized_size >>
 			resize->vol->mft_record_size_bits;
 
-   	for (mref = 0; mref < (MFT_REF)nr_mft_records; mref++)
+	for (mref = 0; mref < (MFT_REF)nr_mft_records; mref++)
 		relocate_inode(resize, mref, 0);
 
-	while(1) {
+	while (1) {
 		highest_vcn = resize->mft_highest_vcn;
 		mref = nr_mft_records;
 		do {
@@ -1769,11 +1740,10 @@ static void relocate_inodes(ntfs_resize_t *resize)
 
 		if (highest_vcn == resize->mft_highest_vcn)
 			err_exit("Sanity check failed! Highest_vcn = %lld. "
-				 "Please report!", highest_vcn);
+				 "Please report!\n", highest_vcn);
 	}
 done:
-	if (resize->mrec)
-		free(resize->mrec);
+	free(resize->mrec);
 }
 
 static void print_hint(ntfs_volume *vol, const char *s, struct llcn_t llcn)
@@ -1803,12 +1773,12 @@ static void advise_on_resize(ntfs_resize_t *resize)
 	if (opt.verbose) {
 		printf("Estimating smallest shrunken size supported ...\n");
 		printf("File feature         Last used at      By inode\n");
-		print_hint(vol, "$MFT", 	resize->last_mft);
+		print_hint(vol, "$MFT",		resize->last_mft);
 		print_hint(vol, "Multi-Record", resize->last_multi_mft);
-		print_hint(vol, "$MFTMirr", 	resize->last_mftmir);
-		print_hint(vol, "Compressed", 	resize->last_compressed);
-		print_hint(vol, "Sparse", 	resize->last_sparse);
-		print_hint(vol, "Ordinary", 	resize->last_lcn);
+		print_hint(vol, "$MFTMirr",	resize->last_mftmir);
+		print_hint(vol, "Compressed",	resize->last_compressed);
+		print_hint(vol, "Sparse",	resize->last_sparse);
+		print_hint(vol, "Ordinary",	resize->last_lcn);
 	}
 
 	print_advise(vol, resize->last_unsupp);
@@ -1852,16 +1822,16 @@ static void rl_truncate(runlist **rl, const VCN last_vcn)
 {
 	int len;
 	VCN vcn;
-	
+
 	len = rl_items(*rl) - 1;
 	if (len <= 0)
 		err_exit("rl_truncate: bad runlist length: %d\n", len);
 
 	vcn = (*rl)[len].vcn;
-	
+
 	if (vcn < last_vcn)
 		rl_expand(rl, last_vcn);
-	
+
 	else if (vcn > last_vcn)
 		if (ntfs_rl_truncate(rl, last_vcn) == -1)
 			perr_exit("ntfs_rl_truncate");
@@ -1901,12 +1871,12 @@ static void truncate_badclust_bad_attr(ntfs_resize_t *resize)
 
 	if (!(rl_bad = ntfs_mapping_pairs_decompress(vol, a, NULL)))
 		perr_exit("ntfs_mapping_pairs_decompress");
-	
+
 	rl_truncate(&rl_bad, nr_clusters);
 
-	a->highest_vcn = cpu_to_le64(nr_clusters - 1LL);
-	a->allocated_size = cpu_to_le64(nr_clusters * vol->cluster_size);
-	a->data_size = cpu_to_le64(nr_clusters * vol->cluster_size);
+	a->highest_vcn = cpu_to_sle64(nr_clusters - 1LL);
+	a->allocated_size = cpu_to_sle64(nr_clusters * vol->cluster_size);
+	a->data_size = cpu_to_sle64(nr_clusters * vol->cluster_size);
 
 	replace_attribute_runlist(vol, resize->ctx, rl_bad);
 
@@ -1948,7 +1918,7 @@ static void realloc_lcn_bitmap(ntfs_resize_t *resize, s64 bm_bsize)
 
 	if (!(tmp = realloc(resize->lcn_bitmap.bm, bm_bsize)))
 		perr_exit("realloc");
-	
+
 	resize->lcn_bitmap.bm = tmp;
 	resize->lcn_bitmap.size = bm_bsize;
 	bitmap_file_data_fixup(resize->new_volume_size, &resize->lcn_bitmap);
@@ -1981,10 +1951,10 @@ static void truncate_bitmap_data_attr(ntfs_resize_t *resize)
 		realloc_bitmap_data_attr(resize, &rl, nr_bm_clusters);
 	}
 
-	a->highest_vcn = cpu_to_le64(nr_bm_clusters - 1LL);
-	a->allocated_size = cpu_to_le64(nr_bm_clusters * vol->cluster_size);
-	a->data_size = cpu_to_le64(bm_bsize);
-	a->initialized_size = cpu_to_le64(bm_bsize);
+	a->highest_vcn = cpu_to_sle64(nr_bm_clusters - 1LL);
+	a->allocated_size = cpu_to_sle64(nr_bm_clusters * vol->cluster_size);
+	a->data_size = cpu_to_sle64(bm_bsize);
+	a->initialized_size = cpu_to_sle64(bm_bsize);
 
 	replace_attribute_runlist(vol, resize->ctx, rl);
 
@@ -2016,7 +1986,7 @@ static void lookup_data_attr(ntfs_volume *vol,
 			     ntfs_attr_search_ctx **ctx)
 {
 	ntfs_inode *ni;
-	ntfschar *ustr = NULL;
+	ntfschar *ustr;
 	int len = 0;
 
 	if (!(ni = ntfs_inode_open(vol, mref)))
@@ -2025,14 +1995,77 @@ static void lookup_data_attr(ntfs_volume *vol,
 	if (!(*ctx = attr_get_search_ctx(ni, NULL)))
 		exit(1);
 
-	if (str2unicode(aname, &ustr, &len) == -1)
+	if ((ustr = ntfs_str2ucs(aname, &len)) == NULL) {
+		perr_printf("Couldn't convert '%s' to Unicode", aname);
 		exit(1);
+	}
 
 	if (ntfs_attr_lookup(AT_DATA, ustr, len, 0, 0, NULL, 0, *ctx))
 		perr_exit("ntfs_lookup_attr");
 
-	if (ustr != AT_UNNAMED)
-		free(ustr);
+	ntfs_ucsfree(ustr);
+}
+
+static int check_bad_sectors(ntfs_volume *vol)
+{
+	ntfs_attr_search_ctx *ctx;
+	ntfs_inode *base_ni;
+	runlist *rl;
+	s64 i, badclusters = 0;
+
+	ntfs_log_verbose("Checking for bad sectors ...\n");
+
+	lookup_data_attr(vol, FILE_BadClus, "$Bad", &ctx);
+
+	base_ni = ctx->base_ntfs_ino;
+	if (!base_ni)
+		base_ni = ctx->ntfs_ino;
+
+	if (NInoAttrList(base_ni)) {
+		err_printf("Hopelessly many bad sectors has been detected!\n");
+		printf("%s", many_bad_sectors_msg);
+		exit(1);
+	}
+
+	if (!ctx->attr->non_resident)
+		err_exit("Resident attribute in $BadClust! Please report to "
+			 "%s\n", NTFS_DEV_LIST);
+	/*
+	 * FIXME: The below would be partial for non-base records in the
+	 * not yet supported multi-record case. Alternatively use audited
+	 * ntfs_attr_truncate after an umount & mount.
+	 */
+	if (!(rl = ntfs_mapping_pairs_decompress(vol, ctx->attr, NULL)))
+		perr_exit("Decompressing $BadClust:$Bad mapping pairs failed");
+
+	for (i = 0; rl[i].length; i++) {
+		/* CHECKME: LCN_RL_NOT_MAPPED check isn't needed */
+		if (rl[i].lcn == LCN_HOLE || rl[i].lcn == LCN_RL_NOT_MAPPED)
+			continue;
+
+		badclusters += rl[i].length;
+		ntfs_log_verbose("Bad cluster: %#8llx - %#llx    (%lld)\n",
+				 rl[i].lcn, rl[i].lcn + rl[i].length - 1,
+				 rl[i].length);
+	}
+
+	if (badclusters) {
+		printf("%sThis software has detected that the disk has at least"
+		       " %lld bad sector%s.\n",
+		       !opt.badsectors ? NERR_PREFIX : "WARNING: ",
+		       badclusters, badclusters - 1 ? "s" : "");
+		if (!opt.badsectors) {
+			printf("%s", bad_sectors_warning_msg);
+			exit(1);
+		} else
+			printf("WARNING: Bad sectors can cause reliability "
+			       "problems and massive data loss!!!\n");
+	}
+
+	free(rl);
+	ntfs_attr_put_search_ctx(ctx);
+
+	return badclusters;
 }
 
 /**
@@ -2086,7 +2119,8 @@ static int setup_lcn_bitmap(struct bitmap *bm, s64 nr_clusters)
 	/* Determine lcn bitmap byte size and allocate it. */
 	bm->size = rounded_up_division(nr_clusters, 8);
 
-	if (!(bm->bm = (unsigned char *)calloc(1, bm->size)))
+	bm->bm = ntfs_calloc(bm->size);
+	if (!bm->bm)
 		return -1;
 
 	bitmap_file_data_fixup(nr_clusters, bm);
@@ -2119,7 +2153,7 @@ static void update_bootsector(ntfs_resize_t *r)
 		r->progress.flags |= NTFS_PROGBAR_SUPPRESS;
 		copy_clusters(r, r->mftmir_rl.lcn, r->mftmir_old,
 			      r->mftmir_rl.length);
-		bs.mftmirr_lcn = cpu_to_le64(r->mftmir_rl.lcn);
+		bs.mftmirr_lcn = cpu_to_sle64(r->mftmir_rl.lcn);
 		r->progress.flags &= ~NTFS_PROGBAR_SUPPRESS;
 	}
 
@@ -2204,9 +2238,12 @@ static ntfs_volume *mount_volume(void)
 			err_exit("Device '%s' is mounted. "
 				 "You must 'umount' it first.\n", opt.volume);
 	}
-
-	if (!(vol = ntfs_mount(opt.volume, opt.ro_flag))) {
-
+	/*
+	 * Pass NTFS_MNT_FORENSIC so that the mount process does not modify the
+	 * volume at all.  We will do the logfile emptying and dirty setting
+	 * later if needed.
+	 */
+	if (!(vol = ntfs_mount(opt.volume, opt.ro_flag | NTFS_MNT_FORENSIC))) {
 		int err = errno;
 
 		perr_printf("Opening '%s' as NTFS failed", opt.volume);
@@ -2216,12 +2253,14 @@ static ntfs_volume *mount_volume(void)
 			printf("%s", corrupt_volume_msg);
 		else if (err == EPERM)
 			printf("%s", hibernated_volume_msg);
-        	else if (err == EOPNOTSUPP)
+		else if (err == EOPNOTSUPP)
 			printf("%s", unclean_journal_msg);
+		else if (err == EBUSY)
+			printf("%s", opened_volume_msg);
 		exit(1);
 	}
 
-	if (vol->flags & VOLUME_IS_DIRTY)
+	if (NVolWasDirty(vol))
 		if (opt.force-- <= 0)
 			err_exit("Volume is scheduled for check.\nRun chkdsk /f"
 				 " and please try again, or see option -f.\n");
@@ -2251,31 +2290,21 @@ static ntfs_volume *mount_volume(void)
  */
 static void prepare_volume_fixup(ntfs_volume *vol)
 {
-	u16 flags;
-
-	flags = vol->flags | VOLUME_IS_DIRTY;
-	if (vol->major_ver >= 2)
-		flags |= VOLUME_MOUNTED_ON_NT4;
-
-	printf("Schedule chkdsk for NTFS consistency check at Windows "
-		"boot time ...\n");
-
-	if (ntfs_volume_write_flags(vol, flags))
-		perr_exit("Failed to set $Volume dirty");
-
+	printf("Schedule chkdsk for NTFS consistency check at Windows boot "
+			"time ...\n");
+	vol->flags |= VOLUME_IS_DIRTY;
+	if (ntfs_volume_write_flags(vol, vol->flags))
+		perr_exit("Failed to set the volume dirty");
+	NVolSetWasDirty(vol);
 	if (vol->dev->d_ops->sync(vol->dev) == -1)
 		perr_exit("Failed to sync device");
-
 	printf("Resetting $LogFile ... (this might take a while)\n");
-
 	if (ntfs_logfile_reset(vol))
 		perr_exit("Failed to reset $LogFile");
-
 	if (vol->dev->d_ops->sync(vol->dev) == -1)
 		perr_exit("Failed to sync device");
 }
 
-
 static void set_disk_usage_constraint(ntfs_resize_t *resize)
 {
 	/* last lcn for a filled up volume (no empty space) */
@@ -2289,17 +2318,6 @@ static void check_resize_constraints(ntfs_resize_t *resize)
 {
 	s64 new_size = resize->new_volume_size;
 
-	if (resize->badclusters) {
-		printf("%sThe NTFS volume has at least %lld bad sector%s.\n",
-		       !opt.badsectors ? NERR_PREFIX : "",
-		       resize->badclusters,
-		       resize->badclusters  - 1 ? "s" : "");
-		if (!opt.badsectors) {
-			printf("%s", bad_sectors_warning_msg);
-			exit(1);
-		}
-	}
-		
 	/* FIXME: resize.shrink true also if only -i is used */
 	if (!resize->shrink)
 		return;
@@ -2326,6 +2344,31 @@ static void check_resize_constraints(ntfs_resize_t *resize)
 	print_num_of_relocations(resize);
 }
 
+static void check_cluster_allocation(ntfs_volume *vol, ntfsck_t *fsck)
+{
+	memset(fsck, 0, sizeof(ntfsck_t));
+
+	if (opt.show_progress)
+		fsck->flags |= NTFSCK_PROGBAR;
+
+	if (setup_lcn_bitmap(&fsck->lcn_bitmap, vol->nr_clusters) != 0)
+		perr_exit("Failed to setup allocation bitmap");
+	if (build_allocation_bitmap(vol, fsck) != 0)
+		exit(1);
+	if (fsck->outsider || fsck->multi_ref) {
+		err_printf("Filesystem check failed!\n");
+		if (fsck->outsider)
+			err_printf("%d clusters are referenced outside "
+				   "of the volume.\n", fsck->outsider);
+		if (fsck->multi_ref)
+			err_printf("%d clusters are referenced multiply"
+				   " times.\n", fsck->multi_ref);
+		printf("%s", corrupt_volume_msg);
+		exit(1);
+	}
+
+	compare_bitmaps(vol, &fsck->lcn_bitmap);
+}
 
 int main(int argc, char **argv)
 {
@@ -2335,6 +2378,8 @@ int main(int argc, char **argv)
 	s64 device_size;        /* in bytes */
 	ntfs_volume *vol;
 
+	ntfs_log_set_handler(ntfs_log_handler_outerr);
+
 	printf("%s v%s (libntfs %s)\n", EXEC_NAME, VERSION,
 			ntfs_libntfs_version());
 
@@ -2343,7 +2388,7 @@ int main(int argc, char **argv)
 
 	utils_set_locale();
 
-	if ((vol = mount_volume()) == NULL)
+	if (!(vol = mount_volume()))
 		err_exit("Couldn't open volume '%s'!\n", opt.volume);
 
 	device_size = ntfs_device_size_get(vol->dev, vol->sector_size);
@@ -2384,47 +2429,34 @@ int main(int argc, char **argv)
 		exit(0);
 	}
 
-	memset(&fsck, 0, sizeof(fsck));
-	if (opt.show_progress)
-		fsck.flags |= NTFSCK_PROGBAR;
-
-	if (setup_lcn_bitmap(&fsck.lcn_bitmap, vol->nr_clusters) != 0)
-		perr_exit("Failed to setup allocation bitmap");
-	if (build_allocation_bitmap(vol, &fsck) != 0)
-		exit(1);
-	if (fsck.outsider || fsck.multi_ref) {
-		err_printf("Filesystem check failed!\n");
-		if (fsck.outsider)
-			err_printf("%d clusters are referenced outside "
-				   "of the volume.\n", fsck.outsider);
-		if (fsck.multi_ref)
-			err_printf("%d clusters are referenced multiply"
-				   " times.\n", fsck.multi_ref);
-		printf("%s", corrupt_volume_msg);
-		exit(1);
-	}
-	compare_bitmaps(vol, &fsck.lcn_bitmap);
-
-	print_disk_usage(vol, fsck.inuse);
-
 	memset(&resize, 0, sizeof(resize));
-	resize.new_volume_size = new_size;
-	resize.inuse = fsck.inuse;
-	resize.lcn_bitmap = fsck.lcn_bitmap;
 	resize.vol = vol;
-	if (opt.show_progress)
-		resize.progress.flags |= NTFS_PROGBAR;
-
+	resize.new_volume_size = new_size;
 	/* This is also true if --info was used w/o --size (new_size = 0) */
 	if (new_size < vol->nr_clusters)
 		resize.shrink = 1;
+	if (opt.show_progress)
+		resize.progress.flags |= NTFS_PROGBAR;
+	/*
+	 * Checking and __reporting__ of bad sectors must be done before cluster
+	 * allocation check because chkdsk doesn't fix $Bitmap's w/ bad sectors
+	 * thus users would (were) quite confused why chkdsk doesn't work.
+	 */
+	resize.badclusters = check_bad_sectors(vol);
+
+	check_cluster_allocation(vol, &fsck);
+
+	print_disk_usage(vol, fsck.inuse);
+
+	resize.inuse = fsck.inuse;
+	resize.lcn_bitmap = fsck.lcn_bitmap;
 
 	set_resize_constraints(&resize);
 	set_disk_usage_constraint(&resize);
 	check_resize_constraints(&resize);
 
 	if (opt.info) {
-	       	advise_on_resize(&resize);
+		advise_on_resize(&resize);
 		exit(0);
 	}
 
