@@ -2,6 +2,7 @@
  * ntfsundelete - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2002 Richard Russon
+ * Copyright (c) 2007 Yura Pakhuchiy
  *
  * This utility will recover deleted files from an NTFS volume.
  *
@@ -57,7 +58,6 @@ struct options {
 	s64		 mft_begin;	/* Range for mft copy */
 	s64		 mft_end;
 	char		 fillbyte;	/* Use for unrecoverable sections */
-	char		 padding[7];	/* Unused: padding to 64 bit. */
 };
 
 struct filename {
@@ -73,9 +73,8 @@ struct filename {
 	time_t		 date_r;	/*	read */
 	char		*name;		/* Filename in current locale */
 	FILE_NAME_TYPE_FLAGS name_space;
-	long long	 parent_mref;
+	leMFT_REF	 parent_mref;
 	char		*parent_name;
-	char		 padding[7];	/* Unused: padding to 64 bit. */
 };
 
 struct data {
@@ -93,7 +92,6 @@ struct data {
 	runlist_element *runlist;	/* Decoded data runs */
 	int		 percent;	/* Amount potentially recoverable */
 	void		*data;		/* If resident, a pointer to the data */
-	char		 padding[4];	/* Unused: padding to 64 bit. */
 };
 
 struct ufile {
@@ -102,12 +100,11 @@ struct ufile {
 	struct list_head name;		/* A list of filenames */
 	struct list_head data;		/* A list of data streams */
 	char		*pref_name;	/* Preferred filename */
-	char		*pref_pname;	/* 	     parent filename */
+	char		*pref_pname;	/*	     parent filename */
 	long long	 max_size;	/* Largest size we find */
 	int		 attr_list;	/* MFT record may be one of many */
 	int		 directory;	/* MFT record represents a directory */
 	MFT_RECORD	*mft;		/* Raw MFT record */
-	char		 padding[4];	/* Unused: padding to 64 bit. */
 };
 
 #endif /* _NTFSUNDELETE_H_ */
